@@ -18,6 +18,12 @@
  */
 
 #include "includes.h"
+#include "system/filesys.h"
+#include "smbd/smbd.h"
+#include "../librpc/gen_ndr/lsa.h"
+#include "../libcli/security/security.h"
+#include "../libcli/security/dom_sid.h"
+#include "passdb.h"
 
 #undef DBGC_CLASS
 #define DBGC_CLASS DBGC_VFS
@@ -30,11 +36,11 @@
 
 #define MAXSIZE 2048
 
-extern const DOM_SID global_sid_World;
-extern const DOM_SID global_sid_Builtin_Administrators;
-extern const DOM_SID global_sid_Builtin_Backup_Operators;
-extern const DOM_SID global_sid_Authenticated_Users;
-extern const DOM_SID global_sid_NULL;
+extern const struct dom_sid global_sid_World;
+extern const struct dom_sid global_sid_Builtin_Administrators;
+extern const struct dom_sid global_sid_Builtin_Backup_Operators;
+extern const struct dom_sid global_sid_Authenticated_Users;
+extern const struct dom_sid global_sid_NULL;
 
 static char space_replacement = '%';
 
@@ -46,7 +52,7 @@ extern int afs_syscall(int, char *, int, char *, int);
 struct afs_ace {
 	bool positive;
 	char *name;
-	DOM_SID sid;
+	struct dom_sid sid;
 	enum lsa_SidType type;
 	uint32 rights;
 	struct afs_ace *next;
@@ -108,7 +114,7 @@ static struct afs_ace *new_afs_ace(TALLOC_CTX *mem_ctx,
 				   bool positive,
 				   const char *name, uint32 rights)
 {
-	DOM_SID sid;
+	struct dom_sid sid;
 	enum lsa_SidType type;
 	struct afs_ace *result;
 
@@ -417,7 +423,7 @@ static void split_afs_acl(struct afs_acl *acl,
 static bool same_principal(struct afs_ace *x, struct afs_ace *y)
 {
 	return ( (x->positive == y->positive) &&
-		 (sid_compare(&x->sid, &y->sid) == 0) );
+		 (dom_sid_compare(&x->sid, &y->sid) == 0) );
 }
 
 static void merge_afs_acls(struct afs_acl *dir_acl,
@@ -514,7 +520,7 @@ static struct static_dir_ace_mapping {
 
 	/* FULL inherit only -- counterpart to previous one */
 	{ 0, SEC_ACE_FLAG_OBJECT_INHERIT|SEC_ACE_FLAG_CONTAINER_INHERIT|SEC_ACE_FLAG_INHERIT_ONLY,
-	  PERMS_FULL | GENERIC_RIGHT_WRITE_ACCESS, 127 /* rlidwka */ },
+	  PERMS_FULL | SEC_GENERIC_WRITE, 127 /* rlidwka */ },
 
 	/* CHANGE without inheritance -- in all cases here we also get
 	   the corresponding INHERIT_ONLY ACE in the same ACL */
@@ -522,13 +528,13 @@ static struct static_dir_ace_mapping {
 
 	/* CHANGE inherit only -- counterpart to previous one */
 	{ 0, SEC_ACE_FLAG_OBJECT_INHERIT|SEC_ACE_FLAG_CONTAINER_INHERIT|SEC_ACE_FLAG_INHERIT_ONLY,
-	  PERMS_CHANGE | GENERIC_RIGHT_WRITE_ACCESS, 63 /* rlidwk */ },
+	  PERMS_CHANGE | SEC_GENERIC_WRITE, 63 /* rlidwk */ },
 
 	/* End marker, hopefully there's no afs right 9999 :-) */
 	{ 0, 0, 0, 9999 }
 };
 
-static uint32 nt_to_afs_dir_rights(const char *filename, const SEC_ACE *ace)
+static uint32 nt_to_afs_dir_rights(const char *filename, const struct security_ace *ace)
 {
 	uint32 result = 0;
 	uint32 rights = ace->access_mask;
@@ -569,7 +575,7 @@ static uint32 nt_to_afs_dir_rights(const char *filename, const SEC_ACE *ace)
 	return result;
 }
 
-static uint32 nt_to_afs_file_rights(const char *filename, const SEC_ACE *ace)
+static uint32 nt_to_afs_file_rights(const char *filename, const struct security_ace *ace)
 {
 	uint32 result = 0;
 	uint32 rights = ace->access_mask;
@@ -590,20 +596,20 @@ static size_t afs_to_nt_acl_common(struct afs_acl *afs_acl,
 				   uint32 security_info,
 				   struct security_descriptor **ppdesc)
 {
-	SEC_ACE *nt_ace_list;
-	DOM_SID owner_sid, group_sid;
-	SEC_ACL *psa = NULL;
+	struct security_ace *nt_ace_list;
+	struct dom_sid owner_sid, group_sid;
+	struct security_acl *psa = NULL;
 	int good_aces;
 	size_t sd_size;
 	TALLOC_CTX *mem_ctx = talloc_tos();
 
 	struct afs_ace *afs_ace;
 
-	uid_to_sid(&owner_sid, psbuf->st_uid);
-	gid_to_sid(&group_sid, psbuf->st_gid);
+	uid_to_sid(&owner_sid, psbuf->st_ex_uid);
+	gid_to_sid(&group_sid, psbuf->st_ex_gid);
 
 	if (afs_acl->num_aces) {
-		nt_ace_list = TALLOC_ARRAY(mem_ctx, SEC_ACE, afs_acl->num_aces);
+		nt_ace_list = TALLOC_ARRAY(mem_ctx, struct security_ace, afs_acl->num_aces);
 
 		if (nt_ace_list == NULL)
 			return 0;
@@ -626,7 +632,7 @@ static size_t afs_to_nt_acl_common(struct afs_acl *afs_acl,
 			continue;
 		}
 
-		if (S_ISDIR(psbuf->st_mode))
+		if (S_ISDIR(psbuf->st_ex_mode))
 			afs_to_nt_dir_rights(afs_ace->rights, &nt_rights,
 					     &flag);
 		else
@@ -642,11 +648,11 @@ static size_t afs_to_nt_acl_common(struct afs_acl *afs_acl,
 	if (psa == NULL)
 		return 0;
 
-	*ppdesc = make_sec_desc(mem_ctx, SEC_DESC_REVISION,
+	*ppdesc = make_sec_desc(mem_ctx, SD_REVISION,
 				SEC_DESC_SELF_RELATIVE,
-				(security_info & OWNER_SECURITY_INFORMATION)
+				(security_info & SECINFO_OWNER)
 				? &owner_sid : NULL,
-				(security_info & GROUP_SECURITY_INFORMATION)
+				(security_info & SECINFO_GROUP)
 				? &group_sid : NULL,
 				NULL, psa, &sd_size);
 
@@ -655,24 +661,24 @@ static size_t afs_to_nt_acl_common(struct afs_acl *afs_acl,
 
 static size_t afs_to_nt_acl(struct afs_acl *afs_acl,
 			    struct connection_struct *conn,
-			    const char *name,
+			    struct smb_filename *smb_fname,
 			    uint32 security_info,
 			    struct security_descriptor **ppdesc)
 {
-	SMB_STRUCT_STAT sbuf;
 	int ret;
 
 	/* Get the stat struct for the owner info. */
 	if (lp_posix_pathnames()) {
-		ret = SMB_VFS_LSTAT(conn, name, &sbuf);
+		ret = SMB_VFS_LSTAT(conn, smb_fname);
 	} else {
-		ret = SMB_VFS_STAT(conn, name, &sbuf);
+		ret = SMB_VFS_STAT(conn, smb_fname);
 	}
 	if (ret == -1) {
 		return 0;
 	}
 
-	return afs_to_nt_acl_common(afs_acl, &sbuf, security_info, ppdesc);
+	return afs_to_nt_acl_common(afs_acl, &smb_fname->st, security_info,
+				    ppdesc);
 }
 
 static size_t afs_fto_nt_acl(struct afs_acl *afs_acl,
@@ -682,7 +688,7 @@ static size_t afs_fto_nt_acl(struct afs_acl *afs_acl,
 {
 	SMB_STRUCT_STAT sbuf;
 
-	if (fsp->is_directory || fsp->fh->fd == -1) {
+	if (fsp->fh->fd == -1) {
 		/* Get the stat struct for the owner info. */
 		return afs_to_nt_acl(afs_acl, fsp->conn, fsp->fsp_name,
 				     security_info, ppdesc);
@@ -695,20 +701,20 @@ static size_t afs_fto_nt_acl(struct afs_acl *afs_acl,
 	return afs_to_nt_acl_common(afs_acl, &sbuf, security_info, ppdesc);
 }
 
-static bool mappable_sid(const DOM_SID *sid)
+static bool mappable_sid(const struct dom_sid *sid)
 {
-	DOM_SID domain_sid;
+	struct dom_sid domain_sid;
 	
-	if (sid_compare(sid, &global_sid_Builtin_Administrators) == 0)
+	if (dom_sid_compare(sid, &global_sid_Builtin_Administrators) == 0)
 		return True;
 
-	if (sid_compare(sid, &global_sid_World) == 0)
+	if (dom_sid_compare(sid, &global_sid_World) == 0)
 		return True;
 
-	if (sid_compare(sid, &global_sid_Authenticated_Users) == 0)
+	if (dom_sid_compare(sid, &global_sid_Authenticated_Users) == 0)
 		return True;
 
-	if (sid_compare(sid, &global_sid_Builtin_Backup_Operators) == 0)
+	if (dom_sid_compare(sid, &global_sid_Builtin_Backup_Operators) == 0)
 		return True;
 
 	string_to_sid(&domain_sid, "S-1-5-21");
@@ -723,15 +729,15 @@ static bool nt_to_afs_acl(const char *filename,
 			  uint32 security_info_sent,
 			  const struct security_descriptor *psd,
 			  uint32 (*nt_to_afs_rights)(const char *filename,
-						     const SEC_ACE *ace),
+						     const struct security_ace *ace),
 			  struct afs_acl *afs_acl)
 {
-	const SEC_ACL *dacl;
+	const struct security_acl *dacl;
 	int i;
 
 	/* Currently we *only* look at the dacl */
 
-	if (((security_info_sent & DACL_SECURITY_INFORMATION) == 0) ||
+	if (((security_info_sent & SECINFO_DACL) == 0) ||
 	    (psd->dacl == NULL))
 		return True;
 
@@ -741,7 +747,7 @@ static bool nt_to_afs_acl(const char *filename,
 	dacl = psd->dacl;
 
 	for (i = 0; i < dacl->num_aces; i++) {
-		const SEC_ACE *ace = &(dacl->aces[i]);
+		const struct security_ace *ace = &(dacl->aces[i]);
 		const char *dom_name, *name;
 		enum lsa_SidType name_type;
 		char *p;
@@ -757,22 +763,22 @@ static bool nt_to_afs_acl(const char *filename,
 			continue;
 		}
 
-		if (sid_compare(&ace->trustee,
+		if (dom_sid_compare(&ace->trustee,
 				&global_sid_Builtin_Administrators) == 0) {
 
 			name = "system:administrators";
 
-		} else if (sid_compare(&ace->trustee,
+		} else if (dom_sid_compare(&ace->trustee,
 				       &global_sid_World) == 0) {
 
 			name = "system:anyuser";
 
-		} else if (sid_compare(&ace->trustee,
+		} else if (dom_sid_compare(&ace->trustee,
 				       &global_sid_Authenticated_Users) == 0) {
 
 			name = "system:authuser";
 
-		} else if (sid_compare(&ace->trustee,
+		} else if (dom_sid_compare(&ace->trustee,
 				       &global_sid_Builtin_Backup_Operators)
 			   == 0) {
 
@@ -911,7 +917,7 @@ static NTSTATUS afs_set_nt_acl(vfs_handle_struct *handle, files_struct *fsp,
 	ZERO_STRUCT(dir_acl);
 	ZERO_STRUCT(file_acl);
 
-	name = talloc_strdup(talloc_tos(), fsp->fsp_name);
+	name = talloc_strdup(talloc_tos(), fsp->fsp_name->base_name);
 	if (!name) {
 		return NT_STATUS_NO_MEMORY;
 	}
@@ -931,7 +937,7 @@ static NTSTATUS afs_set_nt_acl(vfs_handle_struct *handle, files_struct *fsp,
 	}
 
 	if (!afs_get_afs_acl(name, &old_afs_acl)) {
-		DEBUG(3, ("Could not get old ACL of %s\n", fsp->fsp_name));
+		DEBUG(3, ("Could not get old ACL of %s\n", fsp_str_dbg(fsp)));
 		goto done;
 	}
 
@@ -947,7 +953,8 @@ static NTSTATUS afs_set_nt_acl(vfs_handle_struct *handle, files_struct *fsp,
 		}
 
 		free_afs_acl(&dir_acl);
-		if (!nt_to_afs_acl(fsp->fsp_name, security_info_sent, psd,
+		if (!nt_to_afs_acl(fsp->fsp_name->base_name,
+				   security_info_sent, psd,
 				   nt_to_afs_dir_rights, &dir_acl))
 			goto done;
 	} else {
@@ -962,7 +969,8 @@ static NTSTATUS afs_set_nt_acl(vfs_handle_struct *handle, files_struct *fsp,
 		}
 
 		free_afs_acl(&file_acl);
-		if (!nt_to_afs_acl(fsp->fsp_name, security_info_sent, psd,
+		if (!nt_to_afs_acl(fsp->fsp_name->base_name,
+				   security_info_sent, psd,
 				   nt_to_afs_file_rights, &file_acl))
 			goto done;
 	}
@@ -1003,11 +1011,11 @@ static NTSTATUS afsacl_fget_nt_acl(struct vfs_handle_struct *handle,
 	struct afs_acl acl;
 	size_t sd_size;
 
-	DEBUG(5, ("afsacl_fget_nt_acl: %s\n", fsp->fsp_name));
+	DEBUG(5, ("afsacl_fget_nt_acl: %s\n", fsp_str_dbg(fsp)));
 
 	sidpts = lp_parm_bool(SNUM(fsp->conn), "afsacl", "sidpts", False);
 
-	if (!afs_get_afs_acl(fsp->fsp_name, &acl)) {
+	if (!afs_get_afs_acl(fsp->fsp_name->base_name, &acl)) {
 		return NT_STATUS_ACCESS_DENIED;
 	}
 
@@ -1024,6 +1032,8 @@ static NTSTATUS afsacl_get_nt_acl(struct vfs_handle_struct *handle,
 {
 	struct afs_acl acl;
 	size_t sd_size;
+	struct smb_filename *smb_fname = NULL;
+	NTSTATUS status;
 
 	DEBUG(5, ("afsacl_get_nt_acl: %s\n", name));
 
@@ -1033,8 +1043,16 @@ static NTSTATUS afsacl_get_nt_acl(struct vfs_handle_struct *handle,
 		return NT_STATUS_ACCESS_DENIED;
 	}
 
-	sd_size = afs_to_nt_acl(&acl, handle->conn, name, security_info,
+	status = create_synthetic_smb_fname(talloc_tos(), name, NULL, NULL,
+					    &smb_fname);
+	if (!NT_STATUS_IS_OK(status)) {
+		free_afs_acl(&acl);
+		return status;
+	}
+
+	sd_size = afs_to_nt_acl(&acl, handle->conn, smb_fname, security_info,
 				ppdesc);
+	TALLOC_FREE(smb_fname);
 
 	free_afs_acl(&acl);
 
@@ -1044,7 +1062,7 @@ static NTSTATUS afsacl_get_nt_acl(struct vfs_handle_struct *handle,
 NTSTATUS afsacl_fset_nt_acl(vfs_handle_struct *handle,
 			 files_struct *fsp,
 			 uint32 security_info_sent,
-			 const SEC_DESC *psd)
+			 const struct security_descriptor *psd)
 {
 	return afs_set_nt_acl(handle, fsp, security_info_sent, psd);
 }
@@ -1053,33 +1071,31 @@ static int afsacl_connect(vfs_handle_struct *handle,
 			  const char *service, 
 			  const char *user)
 {
-			const char *spc;
+	const char *spc;
+	int ret = SMB_VFS_NEXT_CONNECT(handle, service, user);
+
+	if (ret < 0) {
+		return ret;
+	}
 
 	spc = lp_parm_const_string(SNUM(handle->conn), "afsacl", "space", "%");
 
 	if (spc != NULL)
 		space_replacement = spc[0];
-	
-	return SMB_VFS_NEXT_CONNECT(handle, service, user);
+
+	return 0;
 }
 
-/* VFS operations structure */
-
-static vfs_op_tuple afsacl_ops[] = {	
-	{SMB_VFS_OP(afsacl_connect), SMB_VFS_OP_CONNECT,
-	 SMB_VFS_LAYER_TRANSPARENT},
-	{SMB_VFS_OP(afsacl_fget_nt_acl), SMB_VFS_OP_FGET_NT_ACL,
-	 SMB_VFS_LAYER_TRANSPARENT},
-	{SMB_VFS_OP(afsacl_get_nt_acl), SMB_VFS_OP_GET_NT_ACL,
-	 SMB_VFS_LAYER_TRANSPARENT},
-	{SMB_VFS_OP(afsacl_fset_nt_acl), SMB_VFS_OP_FSET_NT_ACL,
-	 SMB_VFS_LAYER_TRANSPARENT},
-	{SMB_VFS_OP(NULL), SMB_VFS_OP_NOOP, SMB_VFS_LAYER_NOOP}
+static struct vfs_fn_pointers vfs_afsacl_fns = {
+	.connect_fn = afsacl_connect,
+	.fget_nt_acl = afsacl_fget_nt_acl,
+	.get_nt_acl = afsacl_get_nt_acl,
+	.fset_nt_acl = afsacl_fset_nt_acl
 };
 
 NTSTATUS vfs_afsacl_init(void);
 NTSTATUS vfs_afsacl_init(void)
 {
 	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "afsacl",
-				afsacl_ops);
+				&vfs_afsacl_fns);
 }
